@@ -1,23 +1,46 @@
-# -*- coding: utf-8 -*-
+# --------------------------------------------------------------------
+# SPDX-License-Identifier: AGPL-3.0-or-later
+# © Copyright 2008-2024 José Manuel Rodríguez de la Rosa and contributors.
+# See the file CONTRIBUTORS.md for copyright details.
+# See https://www.gnu.org/licenses/agpl-3.0.html for details.
+# --------------------------------------------------------------------
+
 from __future__ import annotations
 
-from typing import Final, Iterable, Iterator, List
+from collections.abc import Iterable, Iterator, Sequence
+from typing import TYPE_CHECKING, Final
 
-import src.api.config
-import src.arch.z80.backend.common
+from src.api import errmsg
+from src.api.config import OPTIONS
 from src.api.debug import __DEBUG__
-from src.api.utils import first
-from src.arch.z80.optimizer import helpers
-from src.arch.z80.optimizer.common import JUMP_LABELS, LABELS
-from src.arch.z80.optimizer.cpustate import CPUState
-from src.arch.z80.optimizer.helpers import ALL_REGS
-from src.arch.z80.optimizer.labelinfo import LabelInfo
-from src.arch.z80.optimizer.memcell import MemCell
-from src.arch.z80.optimizer.patterns import RE_ID_OR_NUMBER
+from src.api.utils import flatten_list, sfirst
+from src.arch.z80.backend.common import ASMS
 from src.arch.z80.peephole import evaluator
+from src.arch.z80.peephole.engine import OptPattern
+from src.arch.z80.peephole.evaluator import FN
+
+from .cpustate import CPUState
+from .helpers import (
+    ALL_REGS,
+    dict_intersection,
+    idx_args,
+    is_16bit_oper_register,
+    new_tmp_val,
+    simplify_asm_args,
+    single_registers,
+)
+from .labelinfo import LabelInfo
+from .labels_dict import LabelsDict
+from .memcell import MemCell
+from .patterns import RE_ID_OR_NUMBER
+
+if TYPE_CHECKING:
+    from .main import Optimizer
+
+__all__ = "BasicBlock", "DummyBasicBlock"
 
 
-class BasicBlock(Iterable[MemCell]):
+class BasicBlock(Sequence[MemCell]):
     """A Class describing a basic block"""
 
     __UNIQUE_ID = 0
@@ -27,9 +50,10 @@ class BasicBlock(Iterable[MemCell]):
         cls.__UNIQUE_ID += 1
         return super().__new__(cls)
 
-    def __init__(self, memory: Iterable[str]):
+    def __init__(self, memory: Iterable[str], optimizer: Optimizer) -> None:
         """Initializes the internal array of instructions."""
-        self.mem: List[MemCell] = []
+        self.optimizer = optimizer
+        self.mem: list[MemCell] = []
         self.next: BasicBlock | None = None  # Which (if any) basic block follows this one in memory
         self.prev: BasicBlock | None = None  # Which (if any) basic block precedes to this one in the code
         self.lock = False  # True if this block is being accessed by other subroutine
@@ -37,14 +61,13 @@ class BasicBlock(Iterable[MemCell]):
         self.goes_to: set[BasicBlock] = set()  # A list/tuple of possible block to jump from here
         self.modified = False  # True if something has been changed during optimization
         self.called_by: set[BasicBlock] = set()
-        self.label_goes = []
         self.ignored = False  # True if this block can be ignored (it's useless)
         self.id: Final[int] = BasicBlock.__UNIQUE_ID
         self._bytes = None
         self._sizeof = None
         self._max_tstates = None
         self.optimized = False  # True if this block was already optimized
-        self.code = memory
+        self.code = list(memory)
         self.cpu = CPUState()
 
     def __hash__(self) -> int:
@@ -57,9 +80,9 @@ class BasicBlock(Iterable[MemCell]):
         return "\n".join(x for x in self.code)
 
     def __repr__(self) -> str:
-        return "<{}: id: {}, len: {}>".format(self.__class__.__name__, self.id, len(self))
+        return f"<{self.__class__.__name__}: id: {self.id}, len: {len(self)}>"
 
-    def __getitem__(self, key) -> MemCell | list[MemCell]:
+    def __getitem__(self, key):
         return self.mem[key]
 
     def __setitem__(self, key, value: MemCell):
@@ -72,31 +95,36 @@ class BasicBlock(Iterable[MemCell]):
         for mem in self.mem:
             yield mem
 
+    @property
+    def jump_labels(self) -> set[str]:
+        return self.optimizer.JUMP_LABELS
+
+    @property
+    def opt_labels(self) -> LabelsDict:
+        return self.optimizer.LABELS
+
     def pop(self, i: int) -> MemCell:
         self._bytes = None
         self._sizeof = None
         self._max_tstates = None
         return self.mem.pop(i)
 
-    def insert(self, i: int, value: str):
-        memcell = MemCell(value, i)
-        self.mem.insert(i, memcell)
-        self._bytes = None
-        self._sizeof = None
-        self._max_tstates = None
-
     @property
-    def code(self) -> List[str]:
+    def code(self) -> list[str]:
         return [x.code for x in self.mem]
 
     @code.setter
     def code(self, value: Iterable[str]):
+        self._set_code(value)
+
+    def _set_code(self, value: Iterable[str]) -> None:
         assert isinstance(value, Iterable)
-        assert all(isinstance(x, str) for x in value)
+        mems = tuple(value)
+        assert all(isinstance(x, str) for x in mems)
         if self.clean_asm_args:
-            self.mem = [MemCell(helpers.simplify_asm_args(asm), i) for i, asm in enumerate(value)]
+            self.mem = [MemCell(simplify_asm_args(asm), i) for i, asm in enumerate(mems)]
         else:
-            self.mem = [MemCell(asm, i) for i, asm in enumerate(value)]
+            self.mem = [MemCell(asm, i) for i, asm in enumerate(mems)]
 
         self._bytes = None
         self._sizeof = None
@@ -108,7 +136,7 @@ class BasicBlock(Iterable[MemCell]):
         if self._bytes is not None:
             return self._bytes
 
-        self._bytes = list(x.bytes for x in self.mem)
+        self._bytes = [x.bytes for x in self.mem]
         return self._bytes
 
     @property
@@ -139,10 +167,10 @@ class BasicBlock(Iterable[MemCell]):
         partitioned or None if there's no such point
         """
         for i, mem in enumerate(self):
-            if i > 0 and mem.is_label and mem.inst in JUMP_LABELS:
+            if i > 0 and mem.is_label and mem.inst in self.jump_labels:
                 return i
 
-            if (mem.is_ender or mem.code in src.arch.z80.backend.common.ASMS) and i < len(self) - 1:
+            if (mem.is_ender or mem.code in ASMS) and i < len(self) - 1:
                 return i + 1
 
         return None
@@ -207,7 +235,7 @@ class BasicBlock(Iterable[MemCell]):
         self.goes_to.add(basic_block)
         basic_block.comes_from.add(self)
 
-    def update_next_block(self):
+    def update_next_block(self) -> None:
         """If the last instruction of this block is a JP, JR or RET (with no
         conditions) then goes_to set contains just a
         single block
@@ -229,14 +257,16 @@ class BasicBlock(Iterable[MemCell]):
         if last.inst == "ret":
             return
 
-        if last.opers[0] not in LABELS.keys():
+        if last.opers[0] not in self.opt_labels:
             __DEBUG__("INFO: %s is not defined. No optimization is done." % last.opers[0], 2)
-            LABELS[last.opers[0]] = LabelInfo(last.opers[0], 0, DummyBasicBlock(ALL_REGS, ALL_REGS))
+            self.opt_labels[last.opers[0]] = LabelInfo(
+                last.opers[0], 0, DummyBasicBlock(ALL_REGS, ALL_REGS, self.optimizer)
+            )
 
-        n_block = LABELS[last.opers[0]].basic_block
+        n_block = self.opt_labels[last.opers[0]].basic_block
         self.add_goes_to(n_block)
 
-    def is_used(self, regs: list[str], i: int, top: int | None = None) -> bool:
+    def is_used(self, regs: Sequence[str], i: int, top: int | None = None) -> bool:
         """Checks whether any of the given regs are required from the given point
         to the end or not.
         """
@@ -247,8 +277,8 @@ class BasicBlock(Iterable[MemCell]):
         top = len(self) if top is None else top + 1
 
         if regs and regs[0][0] == "(" and regs[0][-1] == ")":  # A memory address
-            r16 = helpers.single_registers(regs[0][1:-1]) if helpers.is_16bit_oper_register(regs[0][1:-1]) else []
-            ix = helpers.single_registers(helpers.idx_args(regs[0][1:-1])[0]) if helpers.idx_args(regs[0][1:-1]) else []
+            r16 = single_registers(regs[0][1:-1]) if is_16bit_oper_register(regs[0][1:-1]) else []
+            ix = single_registers(idx_args(regs[0][1:-1])[0]) if idx_args(regs[0][1:-1]) else []  # type: ignore
 
             rr = set(r16 + ix)
             mem_vars = set([] if rr else RE_ID_OR_NUMBER.findall(regs[0]))
@@ -272,7 +302,7 @@ class BasicBlock(Iterable[MemCell]):
 
             return True
 
-        regs = src.api.utils.flatten_list([helpers.single_registers(x) for x in regs])  # make a copy
+        regs = flatten_list([single_registers(x) for x in regs])  # make a copy
         for ii in range(i, top):
             if any(r in regs for r in self.mem[ii].requires):
                 return True
@@ -382,12 +412,12 @@ class BasicBlock(Iterable[MemCell]):
         if not self.comes_from:
             return {}, {}
 
-        regs = first(self.comes_from).cpu.regs
-        mems = first(self.comes_from).cpu.mem
+        regs = sfirst(self.comes_from).cpu.regs
+        mems = dict(sfirst(self.comes_from).cpu.mem)
 
         for blk in list(self.comes_from)[1:]:
-            regs = helpers.dict_intersection(regs, blk.cpu.regs)
-            mems = helpers.dict_intersection(mems, blk.cpu.mem)
+            regs = dict_intersection(regs, blk.cpu.regs)
+            mems = dict_intersection(mems, blk.cpu.mem)
 
         return regs, mems
 
@@ -405,19 +435,25 @@ class BasicBlock(Iterable[MemCell]):
         """Tries to detect peep-hole patterns in this basic block
         and remove them.
         """
+        i: int
+        p: OptPattern
+
         if self.optimized:
             return
 
         changed = True
         code = self.code
-        old_unary = dict(evaluator.Evaluator.UNARY)
-        evaluator.Evaluator.UNARY["GVAL"] = lambda x: self.cpu.get(x)
-        evaluator.Evaluator.UNARY["FLAGVAL"] = lambda x: {
-            "c": str(self.cpu.C) if self.cpu.C is not None else helpers.new_tmp_val(),
-            "z": str(self.cpu.Z) if self.cpu.Z is not None else helpers.new_tmp_val(),
-        }.get(x.lower(), helpers.new_tmp_val())
+        old_unary = dict(evaluator.UNARY)
 
-        if src.api.config.OPTIONS.optimization_level > 3:
+        # monkey-patches some functions in this optimizer level (> 2)
+        evaluator.UNARY[FN.GVAL] = lambda x: self.cpu.get(x)
+        evaluator.UNARY[FN.FLAGVAL] = lambda x: {
+            "c": str(self.cpu.C) if self.cpu.C is not None else new_tmp_val(),
+            "z": str(self.cpu.Z) if self.cpu.Z is not None else new_tmp_val(),
+        }.get(x.lower(), new_tmp_val())
+        evaluator.UNARY[FN.IS_REQUIRED] = lambda x: self.is_used([x], i + len(p.patt))
+
+        if OPTIONS.optimization_level > 3:
             regs, mems = self.guesses_initial_state_from_origin_blocks()
         else:
             regs, mems = {}, {}
@@ -435,7 +471,6 @@ class BasicBlock(Iterable[MemCell]):
                     for var, defline in p.defines:
                         match[var] = defline.expr.eval(match)
 
-                    evaluator.Evaluator.UNARY["IS_REQUIRED"] = lambda x: self.is_used([x], i + len(p.patt))
                     if not p.cond.eval(match):
                         continue
 
@@ -443,8 +478,8 @@ class BasicBlock(Iterable[MemCell]):
                     new_code = list(code)
                     matched = new_code[i : i + len(p.patt)]
                     new_code[i : i + len(p.patt)] = p.template.filter(match)
-                    src.api.errmsg.info("pattern applied [{}:{}]".format("%03i" % p.flag, p.fname))
-                    src.api.debug.__DEBUG__("matched: \n    {}".format("\n    ".join(matched)), level=1)
+                    errmsg.info("pattern applied [{}:{}]".format("%03i" % p.flag, p.fname))
+                    __DEBUG__("matched: \n    {}".format("\n    ".join(matched)), level=1)
                     changed = new_code != code
                     if changed:
                         code = new_code
@@ -457,7 +492,7 @@ class BasicBlock(Iterable[MemCell]):
 
                 self.cpu.execute(asm_line)
 
-        evaluator.Evaluator.UNARY.update(old_unary)  # restore old copy
+        evaluator.UNARY.update(old_unary)  # restore old copy
         self.optimized = True
 
 
@@ -466,8 +501,8 @@ class DummyBasicBlock(BasicBlock):
     about what registers uses an destroys
     """
 
-    def __init__(self, destroys: Iterable[str], requires: Iterable[str]):
-        BasicBlock.__init__(self, [])
+    def __init__(self, destroys: Iterable[str], requires: Iterable[str], optimizer: Optimizer) -> None:
+        BasicBlock.__init__(self, [], optimizer)
         self.__destroys = tuple(destroys)
         self.__requires = set(requires)
         self.code = ["ret"]
@@ -480,140 +515,3 @@ class DummyBasicBlock(BasicBlock):
 
     def is_used(self, regs: Iterable[str], i: int, top: int | None = None) -> bool:
         return len([x for x in regs if x in self.__requires]) > 0
-
-
-def split_block(block: BasicBlock, start_of_new_block: int) -> tuple[BasicBlock, BasicBlock]:
-    assert 0 <= start_of_new_block < len(block), f"Invalid split pos: {start_of_new_block}"
-    new_block = BasicBlock([])
-    new_block.mem = block.mem[start_of_new_block:]
-    block.mem = block.mem[:start_of_new_block]
-
-    new_block.next = block.next
-    block.next = new_block
-    new_block.prev = block
-
-    if new_block.next is not None:
-        new_block.next.prev = new_block
-
-    for blk in list(block.goes_to):
-        block.delete_goes_to(blk)
-        new_block.add_goes_to(blk)
-
-    block.add_goes_to(new_block)
-
-    for i, mem in enumerate(new_block):
-        if mem.is_label and mem.inst in LABELS:
-            LABELS[mem.inst].basic_block = new_block
-            LABELS[mem.inst].position = i
-
-    if block[-1].is_ender:
-        if not block[-1].condition_flag:  # If it's an unconditional jp, jr, call, ret
-            block.delete_goes_to(block.next)
-
-    return block, new_block
-
-
-def compute_calls(basic_blocks: list[BasicBlock], jump_labels: set[str]) -> None:
-    calling_blocks: dict[BasicBlock, BasicBlock] = {}
-
-    # Compute which blocks use jump labels
-    for bb in basic_blocks:
-        if bb[-1].is_ender and (op := bb[-1].branch_arg) in LABELS:
-            LABELS[op].used_by.add(bb)
-
-    # For these blocks, add the referenced block in the goes_to
-    for label in jump_labels:
-        for bb in LABELS[label].used_by:
-            bb.add_goes_to(LABELS[label].basic_block)
-
-    # Annotate which blocks uses call (which should be the last instruction)
-    for bb in basic_blocks:
-        if bb[-1].inst != "call":
-            continue
-
-        op = bb[-1].branch_arg
-        if op in LABELS:
-            LABELS[op].basic_block.called_by.add(bb)
-            calling_blocks[bb] = LABELS[op].basic_block
-
-    # For the annotated blocks, trace their goes_to, and their goes_to from
-    # their goes_to and so on, until ret (unconditional or not) is found, and
-    # save that block in a set for later
-    visited: set[tuple[BasicBlock, BasicBlock]] = set()
-    pending: set[tuple[BasicBlock, BasicBlock]] = set(calling_blocks.items())
-
-    while pending:
-        caller, bb = pending.pop()
-        if (caller, bb) in visited:
-            continue
-
-        visited.add((caller, bb))
-
-        if not bb[-1].is_ender:  # if it does not branch, search in the next block
-            pending.add((caller, bb.next))
-            continue
-
-        if bb[-1].inst in {"ret", "reti", "retn"}:
-            if bb[-1].condition_flag:
-                pending.add((caller, bb.next))
-
-            bb.add_goes_to(caller.next)
-            continue
-
-        if bb[-1].inst in {"call", "rst"}:  # A call from this block
-            if bb[-1].condition_flag:  # if it has conditions, it can return from the next block
-                pending.add((caller, bb.next))
-
-
-def get_jump_labels(main_basic_block: BasicBlock) -> set[str]:
-    """Given the main basic block (which contain the entire program), populate
-    the global JUMP_LABEL set with LABELS used by CALL, JR, JP (i.e JP LABEL0)
-    Also updates the global LABELS index with the pertinent information.
-
-    Any BasicBlock containing a JUMP_LABEL in any position which is not the initial
-    one (0 position) must be split at that point into two basic blocks.
-    """
-    jump_labels: set[str] = set()
-
-    for i, mem in enumerate(main_basic_block):
-        if mem.is_label:
-            LABELS.pop(mem.inst)
-            LABELS[mem.inst] = LabelInfo(
-                label=mem.inst, addr=i, basic_block=main_basic_block, position=i  # Unknown yet
-            )
-            continue
-
-        if not mem.is_ender:
-            continue
-
-        lbl = mem.branch_arg
-        if lbl is None:
-            continue
-
-        jump_labels.add(lbl)
-
-        if lbl not in LABELS:
-            __DEBUG__(f"INFO: {lbl} is not defined. No optimization is done.", 2)
-            LABELS[lbl] = LabelInfo(lbl, 0, DummyBasicBlock(ALL_REGS, ALL_REGS))
-
-    return jump_labels
-
-
-def get_basic_blocks(block: BasicBlock) -> list[BasicBlock]:
-    """If a block is not partitionable, returns a list with the same block.
-    Otherwise, returns a list with the resulting blocks.
-    """
-    result: list[BasicBlock] = [block]
-    JUMP_LABELS.clear()
-    JUMP_LABELS.update(get_jump_labels(block))
-
-    # Split basic blocks per label or branch instruction
-    split_pos = block.get_first_partition_idx()
-    while split_pos is not None:
-        _, block = split_block(block, split_pos)
-        result.append(block)
-        split_pos = block.get_first_partition_idx()
-
-    compute_calls(result, JUMP_LABELS)
-
-    return result
